/**********************************************************************
 * 
 *      Original Author: Eric Boucher
 *      Date:            05/05/2003
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 *      Beamline Controls & Data Acquisition Group
 *      Experimental Facilities Division
 *      Advanced Photon Source
 *      Argonne National Laboratory
 *
 *
 * Channel.java,v 1.2 2003/10/13 13:49:39 boucher Exp
 *
 * Modification Log:
 * 01. 05/07/2003  erb  initial development
 *
 */


package gov.aps.epics.jca;

import java.io.PrintStream;

import gov.aps.epics.jca.dbr.*;

/**
 * The class representing a CA Channel.
 * <P>
 * A Channel is a link between a client (the application) and a CA process variable located on a CA server.
 * All operations between the client and the process variable are handled by this object of this class.
 * <pre>
 * // Access a process variable thru a Channel.
 * 
 * try {
 *
 *   // Create a default context.
 *   Context theContext= Ca.createContext();
 *
 *   // Print some info about the context.
 *   Ca.printInfo(theContext);
 *   System.out.println();
 *
 *   // Create a Channel
 *   Channel theChannel= theContext.createChannel("myProcessVariableName");
 *   theContext.pendIO(1.0);
 *
 *   // Print some info about the channel.
 *   Ca.printInfo(theChannel);
 *   System.out.println();
 * 
 *   // Use the Channel
 *   theChannel.put(45.5);
 *   theContext.pendIO(1.0);
 *
 *   DBR theDBR= theChannel.get();
 *   theContext.pendIO(1.0);
 *
 *   // Print some info about the requested value.
 *   Ca.printInfo(theDBR);
 *   System.out.println();
 *
 *   // Clear the channel and destroy the context.
 *   theChannel.clear();
 *   theContext.destroy();
 *
 * } catch(CaException cae) {
 *   cae.printStackTrace();
 * }
 *
 * </pre>
 * 
 * @see Ca
 * @see Context
 * @see DBR
 **/
abstract public class Channel implements CaConstants {

  /**
   * Returns the context which created this channel.
   */
  abstract public Context getContext();

  /**
   * Clear the ressources used by this channel.
   * No further access should be made to a channel after it has been cleared.
   */
  abstract public void destroy() throws CaException;

  /**
   * Returns the ConnectionListener registered with this Channel. Might be null.
   */
  abstract public ConnectionListener[] getConnectionListeners();

  /**
   * Adds a ConnectionListener which will be notified of the connection state's changes of this Channel.
   *
   * @param l the ConnectionListener to be registered.
   */
  abstract public void addConnectionListener(ConnectionListener l) throws CaException;

  /**
   * Removes a ConnectionListener .
   *
   * @param l the ConnectionListener to be removed.
   */
  abstract public void removeConnectionListener(ConnectionListener l) throws CaException;


  /**
   * Returns the AccessRightsListener registered with this Channel. Might be null.
   */
  abstract public AccessRightsListener[] getAccessRightsListeners();
  
  /**
   * Adds a AccessRightsListener which will be notified of the access rights's changes of this Channel.
   *
   * @param l the ConnectionListener to be registered.
   */
  abstract public void addAccessRightsListener(AccessRightsListener l) throws CaException;

  /**
   * Removes a AccessRightsListener which will be notified of the access rights's changes of this Channel.
   *
   * @param l the ConnectionListener to be removed.
   */
  abstract public void removeAccessRightsListener(AccessRightsListener l) throws CaException;

  /**
   * Returns the name of this Channel.
   *
   * @return the name.
   **/
  abstract public String  getName();

  /**
   * Returns the String representing the field type of this channel.
   *
   * @return the field type String.
   **/
  public String getFieldTypeString() {
    return Ca.getDBRTypeString(getFieldType());
  }

  /**
   * Returns the field type of this Channel.
   *
   * @return one of DBR_XXX type.
   * @see CaConstants
   **/
  abstract public int getFieldType();

  /**
   * Returns the element count of this Channel.
   *
   * @return the element count.
   **/
  abstract public int getElementCount();

  /**
   * Returns the connection state of this channel
   *
   * @return one of CS_XXX constants.
   * @see CaConstants
   **/
  abstract public int getState();
  
  public String getStateString() {
    return Ca.getConnectionStateString(getState());
  }

  /**
   * Returns the Channel's hostname.
   *
   * @return the hostname.
   **/
  abstract public String getHostName();

  /**
   * Returns whether read operations are allowed on this Channel.
   *
   * @return true is read operations are allowed, false otherwise.
   **/
  abstract public boolean getReadAccess();

  /**
   * Returns whether write operations are allowed on this Channel.
   *
   * @return true is write operations are allowed, false otherwise.
   **/
  abstract public boolean getWriteAccess();

  
  /**
   * Writes a value to this Channel.
   *
   * @param value the value.
   */
  public void put(byte value) throws CaException {
    put(new byte[] { value });
  }

  /**
   * Writes a value to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  public void put(byte value, PutListener l) throws CaException {
    put(new byte[] { value }, l);
  }

  /**
   * Writes a value to this Channel.
   *
   * @param value the value.
   */
  public void put(short value) throws CaException {
    put(new short[] { value });
  }

  /**
   * Writes a value to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  public void put(short value, PutListener l) throws CaException {
    put(new short[] { value }, l);
  }
  
  /**
   * Writes a value to this Channel.
   *
   * @param value the value.
   */
  public void put(int value) throws CaException {
    put(new int[] { value });
  }

  /**
   * Writes a value to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  public void put(int value, PutListener l) throws CaException {
    put(new int[] { value }, l);
  }

  /**
   * Writes a value to this Channel.
   *
   * @param value the value.
   */
  public void put(float value) throws CaException {
    put(new float[] { value });
  }

  /**
   * Writes a value to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  public void put(float value, PutListener l) throws CaException {
    put(new float[] { value }, l);
  }

  /**
   * Writes a value to this Channel.
   *
   * @param value the value.
   */
  public void put(double value) throws CaException {
    put(new double[] { value });
  }

  /**
   * Writes a value to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  public void put(double value, PutListener l) throws CaException {
    put(new double[] { value }, l);
  }

  /**
   * Writes a value to this Channel.
   *
   * @param value the value.
   */
  public void put(String value) throws CaException {
    put(new String[] { value });
  }
  
  /**
   * Writes a value to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  public void put(String value, PutListener l) throws CaException {
    put(new String[] { value }, l);
  }




  /**
   * Writes an array to this Channel.
   *
   * @param value the value.
   */
  abstract public void put(byte[] value) throws CaException;
  /**
   * Writes an array to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  abstract public void put(byte[] value, PutListener l) throws CaException;
  
  /**
   * Writes an array to this Channel.
   *
   * @param value the value.
   */
  abstract public void put(short[] value) throws CaException;
  /**
   * Writes an array to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  abstract public void put(short[] value, PutListener l) throws CaException;

  /**
   * Writes an array to this Channel.
   *
   * @param value the value.
   */
  abstract public void put(int[] value) throws CaException;
  /**
   * Writes an array to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  abstract public void put(int[] value, PutListener l) throws CaException;

  /**
   * Writes an array to this Channel.
   *
   * @param value the value.
   */
  abstract public void put(float[] value) throws CaException;
  /**
   * Writes an array to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  abstract public void put(float[] value, PutListener l) throws CaException;

  /**
   * Writes an array to this Channel.
   *
   * @param value the value.
   */
  abstract public void put(double[] value) throws CaException;
  /**
   * Writes an array to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  abstract public void put(double[] value, PutListener l) throws CaException;

  /**
   * Writes an array to this Channel.
   *
   * @param value the value.
   */
  abstract public void put(String[] value) throws CaException;
  /**
   * Writes an array to this Channel with a PutListener to be notified when the operation has been completed.
   *
   * @param value the value.
   */
  abstract public void put(String[] value, PutListener l) throws CaException;


  /**
   * Reads getElementCount() of getFieldType() elements from this Channel.
   *
   */
  public DBR get() throws CaException {
    return get(getFieldType(), getElementCount());
  }

  /**
   * Reads getElementCount() of getFieldType() elements from this Channel 
   * with a GetListener to be notified when the operation has been completed.
   *
   * @param l the GetListener.
   */
  public void get(GetListener l) throws CaException{
    get(getFieldType(), getElementCount(), l);
  }

  /**
   * Reads a specified number of getFieldType() elements from this Channel.
   *
   * @param count the number of element to read.
   */
  public DBR get(int count) throws CaException {
    return get(getFieldType(), count);
  }

  /**
   * Reads a specified number of getFieldType() elements from this Channel 
   * with a GetListener to be notified when the operation has been completed.
   *
   * @param count the number of element to read.
   * @param l the GetListener.
   */
  public void get(int count, GetListener l) throws CaException {
    get(getFieldType(), count, l);
  }
  
  /**
   * Reads a specified number of elements of a specified type from this Channel.
   *
   * @param count the number of element to read.
   */
  abstract public DBR  get(int type, int count) throws CaException;
  /**
   * Reads a specified number of elements of a specified type from this Channel 
   * with a GetListener to be notified when the operation has been completed.
   *
   * @param type the element type.
   * @param count the number of element to read.
   * @param l the GetListener.
   */
  abstract public void get(int type, int count, GetListener l) throws CaException;

  /**
   * Adds a monitor to this Channel with a specified type,count and mask.
   *
   * @param type the monitor element type.
   * @param count the monitor element count.
   * @param l the MonitorListener to be notified.
   * @param mask the mask value indicating when the listener need to be notified.
   *
   * @return the Monitor object representing this monitor.
   *
   */
  abstract public Monitor addMonitorListener(int type, int count, MonitorListener l, int mask) throws CaException;

  public Monitor addMonitorListener(MonitorListener l, int mask) throws CaException {
    return addMonitorListener(getFieldType(), getElementCount(), l, mask);
  }


  /**
   * Prints details information about this Channel to the standard output stream.
   */
  public void printInfo() {
    printInfo(System.out);
  }

  /**
   * Prints details information about this Channel to the specified output stream.
   *
   * @param the output stream.
   */
  public void printInfo(PrintStream out) {
    out.println("CHANNEL  : "+getName());
    out.println("TYPE     : "+getFieldTypeString());
    out.println("COUNT    : "+getElementCount());
    out.println("STATE    : "+getStateString());
    out.println("HOSTNAME : "+getHostName());
    out.println("READ     : "+getReadAccess());
    out.println("WRITE    : "+getWriteAccess());
  }

}




