/**********************************************************************
 * 
 *      Original Author: Eric Boucher
 *      Date:            05/05/2003
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 *      Beamline Controls & Data Acquisition Group
 *      Experimental Facilities Division
 *      Advanced Photon Source
 *      Argonne National Laboratory
 *
 *
 * Context.java,v 1.1.1.1 2003/09/11 14:42:49 boucher Exp
 *
 * Modification Log:
 * 01. 05/07/2003  erb  initial development
 *
 */


package gov.aps.epics.jca;

import gov.aps.epics.jca.jni.*;

import java.util.*;

import java.io.PrintStream;

/**
 * The class representing a CA Context.
 * <P>
 * A Context controls all IO operations and the circuits thru which Channels will be created and connected.
 *
 * Since CA 3.14 an application can create more than one context, thus ehancing IO control and Channels managment.
 * It is also possible for the application to specify how asynchronous callbacks will be provided by a context.
 * <pre>
 * Two differents mode are possible:
 *   ENABLE_PREEMPTIVE_CALLBACK : The callbacks are controled by independant threads. The application 
 *                                does not need to call pendEvent() to process pending event.
 *   DISABLE_PREEMPTIVE_CALLBACK: The callbacks are controled by the application itself which
 *                                need to call pendEvent() to process pending events.
 * 
 * </pre>
 *
 * In earlier CA version, only on context using the DISABLE_PREEMPTIVE_CALLBACK mode is allowed.
 *
 * The Context class also offers two different thread control mode.
 * <pre>
 *   THREAD_CTRL_EXTERNAL: in this mode only the thread that created the Context (and all 
 *                         others that have registered with the Context) are allowed to perform
 *                         operations.
 *   THREAD_CTRL_INTERNAL: in this mode the Context uses its own internal thread to perform
 *                         operations, thus allowing any thread in the application to access the Context.
 *
 * </pre>
 * 
 * For complexity reasons uses THREAD_CTRL_EXTERNAL mode in simple application
 * or in application where you want to control which threads will access the Context.
 *
 *
 * @see Ca
 * @see CaConstants
 */
public abstract class Context implements CaConstants {

  /**
   * Prints detailed information about the Context to the standard output stream.
   */
  public void printInfo() {
    printInfo(System.out);
  }

  /**
   * Prints detailed information about the Context to the specified output stream.
   *
   * @param out the output stream.
   */
  public void printInfo(PrintStream out) {
    out.println("NAME        : "+getName());
    out.println("CALLBACK    : "+getCallbackModeString());
    out.println("THREAD CTRL : "+getThreadCtrlModeString());
    out.println("CLASS       : "+getClass().getName());
  }



  static protected HashMap _contextMap=new HashMap();

  static protected void registerContext(Context c) {
    _contextMap.put(c.getName(), c);
  }

  static protected void unregisterContext(Context c) {
    _contextMap.remove(c.getName());
  }

  static protected Context lookupContext(String name) {
    return (Context) _contextMap.get(name);

  }

  /**
   * Returns all created Contexts.
   */
  static public Context[] getContextList() {
    return (Context[]) _contextMap.values().toArray(new Context[0]);
  }



  public void dispose() {
    try {
      destroy();
    } catch(Throwable th) {
    }
  }


  /**
   * Returns the name of the Context.
   */
  abstract public String getName();

  /**
   * Returns the Context's thread control mode.
   *
   * @see CaConstants
   */
  abstract public int getThreadCtrlMode();

  /**
   * Returns a String representing the Context's thread control mode.
   *
   * @see CaConstants
   */
  public String getThreadCtrlModeString() {
    return Ca.getThreadCtrlModeString(getThreadCtrlMode());
  }
  
  /**
   * Returns the Context's callback mode.
   *
   * @see CaConstants
   */
  abstract public int getCallbackMode();

  /**
   * Returns a String representing the Context's callback mode.
   *
   * @see CaConstants
   */
  public String getCallbackModeString() {
    return Ca.getCallbackModeString(getCallbackMode());
  }

  /**
   * Returns the ContextMessageListener registered with the Context.
   *
   */
  abstract public ContextMessageListener getContextMessageListener();

  /**
   * Sets the ContextMessageListener which will receive ContextMessageEvents from this Context.
   *
   * @param l the ContextMessageListener.
   */
  abstract public void setContextMessageListener(ContextMessageListener l) throws CaException;

  /**
   * Returns the ContextExceptionListener registered with the Context.
   *
   */
  abstract public ContextExceptionListener getContextExceptionListener();
  
  /**
   * Sets the ContextExceptionListener which will receive ContextExceptionEvents from this Context.
   *
   * @param l the ContextExceptionListener.
   */
  abstract public void setContextExceptionListener(ContextExceptionListener l) throws CaException;

  /**
   * Destroys the Context and clear all associated ressources.
   */
  abstract public void destroy() throws CaException;

  /**
   * Create a new Channel using this context to access a process variable.
   *
   * @param name the process variable to connect to.
   * @return the new Channel.
   */   
  public Channel createChannel(String name) throws CaException {
    return createChannel(name, null);
  }

  /**
   * Create a new Channel using this context to access a process variable with a ConnectionListener
   * to receive ConnectionEvent from the newly created Channel.
   *
   * @param name the process variable to connect to.
   * @param l the ConnectionListener.
   * @return the new Channel.
   * @exception CaException is thrown if the Context has been destroyed.
   */   
  abstract public Channel createChannel(String name, ConnectionListener l) throws CaException;

  /**
   * Process all pending IO operations.
   *
   * @param timeout the timeout limit.
   * @exception CaException is thrown if all the IO couldn't be processed.
   * @exception CaException is thrown if the Context has been destroyed.
   */
  abstract public void pendIO(double timeout) throws CaException;

  /** 
   * Tests if there are some pending IO requests.
   */
  abstract public boolean testIO() throws CaException;
  
  /**
   * Process all pending events.
   *
   * @param time the time during which the Context will process events.
   */
  abstract public void pendEvent(double time) throws CaException;

  abstract public void poll() throws CaException;

  abstract public void flushIO() throws CaException; 

  abstract public void attachCurrentThread() throws CaException;
}


