/**********************************************************************
 * 
 *      Original Author: Eric Boucher
 *      Date:            05/05/2003
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 *      Beamline Controls & Data Acquisition Group
 *      Experimental Facilities Division
 *      Advanced Photon Source
 *      Argonne National Laboratory
 *
 *
 * JNIChannel.java,v 1.1.1.1 2003/09/11 14:42:49 boucher Exp
 *
 * Modification Log:
 * 01. 05/07/2003  erb  initial development
 *
 */


package gov.aps.epics.jca.jni;

import gov.aps.epics.jca.*;
import gov.aps.epics.jca.dbr.*;

import java.util.ArrayList;
import java.util.Iterator;


public class JNIChannel extends Channel {
  protected JNIContext _jnicontext;
  protected int        _channelID=0;
  protected String     _name;
  protected ArrayList _cnxListenerList= new ArrayList();
  protected ArrayList _accessListenerList= new ArrayList();

  protected JNIChannel(JNIContext context, String name, ConnectionListener l) {
    _jnicontext= context;
    _name= name;
    if(l!=null) _cnxListenerList.add(l);

  }

  protected void setChannelID(int channelID) {
    if(_channelID!=0) throw new IllegalStateException("ChannelID has already been set");
    _channelID= channelID;
  }

  public Context getContext() {
    assertValidity();
    return _jnicontext;
  }

  public void destroy() throws CaException {
    assertValidity();
    dispose();
  }

  class ConnectionEventDispatcher implements ConnectionListener {
    public void connectionChanged(ConnectionEvent ev) {
      System.out.println("From Connection event dispatcher");
      for(Iterator it= _cnxListenerList.iterator(); it.hasNext();) {
        ((ConnectionListener)it.next()).connectionChanged(ev);
      }
    }
  }

  public ConnectionListener[] getConnectionListeners() {
    assertValidity();
    return (ConnectionListener[]) _cnxListenerList.toArray(new ConnectionListener[0]);
  }

  public void addConnectionListener(ConnectionListener l) throws CaException {
    assertValidity();
    if(l!=null) _cnxListenerList.add(l);

    if(_cnxListenerList.size()==1) _jnicontext.ch_setConnectionListener(this, new ConnectionEventDispatcher());
  }

  public void removeConnectionListener(ConnectionListener l) throws CaException {
    assertValidity();
    if(l!=null) _cnxListenerList.remove(l);

    if(_cnxListenerList.size()==0) _jnicontext.ch_setConnectionListener(this, null);
  }

  class AccessRightsEventDispatcher implements AccessRightsListener {
    public void accessRightsChanged(AccessRightsEvent ev) {
      for(Iterator it= _accessListenerList.iterator(); it.hasNext();) {
        ((AccessRightsListener)it.next()).accessRightsChanged(ev);
      }
    }
  }

  public AccessRightsListener[] getAccessRightsListeners() {
    assertValidity();
    return (AccessRightsListener[]) _accessListenerList.toArray(new AccessRightsListener[0]);
  }
  
  public void addAccessRightsListener(AccessRightsListener l) throws CaException {
    assertValidity();
    if(l!=null) _accessListenerList.add(l);

    if(_accessListenerList.size()==1) _jnicontext.ch_setAccessRightsListener(this, new AccessRightsEventDispatcher());
  }

  public void removeAccessRightsListener(AccessRightsListener l) throws CaException {
    assertValidity();
    if(l!=null) _accessListenerList.remove(l);

    if(_accessListenerList.size()==0) _jnicontext.ch_setAccessRightsListener(this, null);
  }



  /**
   * return the name of the process variable.
   *
   * @exception IllegalOperationException is thrown if the process variable has been disposed.
   **/
  public String  getName() {
    assertValidity();
    return _name;
  }

  /**
   * return the pv's channel access type.
   *
   * @exception IllegalOperationException is thrown if the process variable has been disposed.
   **/
  public int getFieldType() {
    assertValidity();
    return _jnicontext.ch_getFieldType(this);
  }

  /**
   * return the pv's element count.
   *
   * @exception IllegalOperationException is thrown if the process variable has been disposed.
   **/
  public int getElementCount() {
    assertValidity();
    return _jnicontext.ch_getElementCount(this);
  }

  /**
   * return the pv's connection state.
   *
   * @exception IllegalOperationException is thrown if the process variable has been disposed.
   **/
  public int getState() {
    assertValidity();
    return _jnicontext.ch_getState(this);
  }

  /**
   * return the pv's hostname.
   *
   * @exception IllegalOperationException is thrown if the process variable has been disposed.
   **/
  public String getHostName() {
    assertValidity();
    return _jnicontext.ch_getHostName(this);
  }

  /**
   * return the pv's read access right.
   *
   * @exception IllegalOperationException is thrown if the process variable has been disposed.
   **/
  public boolean getReadAccess() {
    assertValidity();
    return _jnicontext.ch_getReadAccess(this);
  }

  /**
   * return the pv's write access right.
   *
   * @exception IllegalOperationException is thrown if the process variable has been disposed.
   **/
  public boolean getWriteAccess() {
    assertValidity();
    return _jnicontext.ch_getWriteAccess(this);
  }


  public void dispose() {
    if(_channelID==0) return;
    try {
    _jnicontext.ch_channelDestroy(this);
    } catch(CaException cae) {
    }
    _jnicontext.unregisterChannel(this);

    _jnicontext=null;
    _name=null;
    _channelID=0;
    _cnxListenerList.clear();
    _accessListenerList.clear();
  }

  public int getChannelID() {
    assertValidity();
    return _channelID;
  }


  public void put(byte[] value) throws CaException {
    put(DBR_BYTE,value.length, value);
  }
  public void put(byte[] value, PutListener l) throws CaException {
    put(DBR_BYTE,value.length, value, l);
  }
  
  public void put(short[] value) throws CaException {
    put(DBR_SHORT,value.length, value);
  }
  public void put(short[] value, PutListener l) throws CaException {
    put(DBR_SHORT,value.length, value, l);
  }

  public void put(int[] value) throws CaException {
    put(DBR_INT,value.length, value);
  }
  public void put(int[] value, PutListener l) throws CaException {
    put(DBR_INT,value.length, value, l);
  }

  public void put(float[] value) throws CaException {
    put(DBR_FLOAT,value.length, value);
  }
  public void put(float[] value, PutListener l) throws CaException {
    put(DBR_FLOAT,value.length, value, l);
  }

  public void put(double[] value) throws CaException {
    put(DBR_DOUBLE,value.length, value);
  }
  public void put(double[] value, PutListener l) throws CaException {
    put(DBR_DOUBLE,value.length, value, l);
  }

  public void put(String[] value) throws CaException {
    put(DBR_STRING,value.length, value);
  }

  public void put(String[] value, PutListener l) throws CaException {
    put(DBR_STRING,value.length, value, l);
  }


  protected void put(int type, int count, Object value) throws CaException {
    assertValidity();
    _jnicontext.ch_arrayPut(type,count,this,value);
  }
  protected void put(int type, int count, Object value, PutListener l) throws CaException {
    assertValidity();
    _jnicontext.ch_arrayPut(type,count,this,value,l);
  }



  public DBR  get(int type, int count) throws CaException {
    assertValidity();
    DBR dbr= DBRFactory.create(type,count);
    _jnicontext.ch_arrayGet(dbr,this);
    return dbr;
  }

  public void get(int type, int count, GetListener l) throws CaException {
    assertValidity();
    _jnicontext.ch_arrayGet(type,count, this, l);
  }



  public Monitor addMonitorListener(int type, int count, MonitorListener l, int mask) throws CaException {
    assertValidity();
    return _jnicontext.ch_addMonitor(type,count,this,l,mask);
  }

  protected void assertValidity() {
    if(_channelID==0) throw new IllegalStateException("Invalid channel");
  }
}




