/**********************************************************************
 * 
 *      Original Author: Eric Boucher
 *      Date:            05/05/2003
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 *      Beamline Controls & Data Acquisition Group
 *      Experimental Facilities Division
 *      Advanced Photon Source
 *      Argonne National Laboratory
 *
 *
 * JNIContextTCE.java,v 1.1.1.1 2003/09/11 14:42:49 boucher Exp
 *
 * Modification Log:
 * 01. 05/07/2003  erb  initial development
 *
 */


package gov.aps.epics.jca.jni;

import gov.aps.epics.jca.*;
import gov.aps.epics.jca.dbr.*;

import java.util.LinkedList;
import java.util.Iterator;


public class JNIContextTCE extends JNIContext {
  static protected int _num=0;

  Thread _thread;
  int    _callback;
  String _name;

  LinkedList _threadList= new LinkedList();


  public JNIContextTCE(String name, int callback) throws CaException {
    
    if(name==null || name.trim().equals("")) {
      do {
        name="JNIContextTCE-"+(_num++);
      } while(lookupContext(name)!=null);
    } else {
      if(lookupContext(name)!=null) throw new IllegalArgumentException("Context "+name+" already exists");
    }
    _name= name;
    

    setCtxtID( ctxt_contextCreate(callback));
  
    setContextMessageListener(new DefaultContextMessageListener());
    setContextExceptionListener(new DefaultContextExceptionListener());
  
    registerContext(this);
  }


  public int ctxt_getThreadCtrlMode() {
    return THREAD_CTRL_EXTERNAL;
  }
  
  public int ctxt_getCallbackMode() {
    return _callback;
  }

  public String ctxt_getName() {
    return _name;
  }

  public int ctxt_contextCreate(int callback) throws CaException {
    if(_thread!=null) throw new IllegalStateException("Context already created");
    _callback=callback;

    _threadList.add(Thread.currentThread());

    return JNI._ctxt_contextCreate(callback);
  }

  public void ctxt_contextDestroy() throws CaException {
    assertThread();
    unregisterContext(this);
    JNI._ctxt_contextDestroy();
  }

  public void ctxt_setMessageListener(ContextMessageListener l) throws CaException {
    assertThread();
    JNIContextMessageCallback cb=null;
    if(l!=null) cb= new JNIContextMessageCallback(this,l);
    JNI._ctxt_setMessageCallback(cb);
  }

  public void ctxt_setExceptionListener(ContextExceptionListener l) throws CaException {
    assertThread();
    JNIContextExceptionCallback cb=null;
    if(l!=null) cb= new JNIContextExceptionCallback(this,l);
    JNI._ctxt_setExceptionCallback(cb);
  }

  public void ctxt_pendIO(double timeout) throws CaException {
    assertThread();
    synchronized(_GetQueue) {
      try {
        JNI._ctxt_pendIO(timeout);
      } catch(CaException cae) {
        throw cae;
      } finally {
        clearGetQueue();
      }
    }
  }

  public boolean  ctxt_testIO() throws CaException {
    assertThread();
    return JNI._ctxt_testIO();
  }

  public void ctxt_pendEvent(double time) throws CaException {
    assertThread();
    synchronized(_GetQueue) {
      try {
        JNI._ctxt_pendEvent(time);
      } catch(CaException cae) {
        throw cae;
      } finally {
        clearGetQueue();
      }
    }
  }

  public void ctxt_poll() throws CaException {
    assertThread();
    JNI._ctxt_poll();
  }

  public void ctxt_flushIO() throws CaException {
    assertThread();
    JNI._ctxt_flushIO();
  }

  public void ctxt_attachThread() throws CaException {
    if(ctxt_getCallbackMode()==DISABLE_PREEMPTIVE_CALLBACK) throw new CaException(ECA_NOTTHREADED);

    JNI._ctxt_attachThread(getCtxtID());
    synchronized(_threadList) {
      _threadList.add(Thread.currentThread());
    }
  }



  public JNIChannel ch_channelCreate(String name, ConnectionListener l) throws CaException {
    assertThread();
    JNIChannel res= new JNIChannel(this,name,l);
    JNIConnectionCallback cb=null;
    if(l!=null) cb= new JNIConnectionCallback(res,l);
    res.setChannelID(JNI._ch_channelCreate(name,cb));
    return res;
  }

  public void ch_channelDestroy(JNIChannel ch) throws CaException {
    assertThread();
    JNI._ch_channelDestroy(ch.getChannelID());
  }

  public void ch_setConnectionListener(JNIChannel ch, ConnectionListener l) throws CaException {
    assertThread();
    JNIConnectionCallback cb= null;
    if(l!=null) cb= new JNIConnectionCallback(ch,l);
    JNI._ch_setConnectionCallback(ch.getChannelID(), cb);
  }
  
  public void ch_setAccessRightsListener(JNIChannel ch, AccessRightsListener l) throws CaException {
    assertThread();
    JNIAccessRightsCallback cb= null;
    if(l!=null) cb= new JNIAccessRightsCallback(ch,l);
    JNI._ch_setAccessRightsCallback(ch.getChannelID(), cb);
  }

  public int ch_getFieldType(JNIChannel ch) {
    assertThread();
    return JNI._ch_getFieldType(ch.getChannelID());
  }

  public int ch_getElementCount(JNIChannel ch) {
    assertThread();
    return JNI._ch_getElementCount(ch.getChannelID());
  }

  public int ch_getState(JNIChannel ch) {
    assertThread();
    return JNI._ch_getState(ch.getChannelID());
  }

  public String ch_getHostName(JNIChannel ch) {
    assertThread();
    return JNI._ch_getHostName(ch.getChannelID());
  }

  public boolean ch_getReadAccess(JNIChannel ch) {
    assertThread();
    return JNI._ch_getReadAccess(ch.getChannelID());
  }

  public boolean ch_getWriteAccess(JNIChannel ch) {
    assertThread();
    return JNI._ch_getWriteAccess(ch.getChannelID());
  }


  public void ch_arrayPut(int type, int count, JNIChannel ch, Object value) throws CaException {
    assertThread();
    int dbrid= JNI._dbr_create(type,count);
    JNI._dbr_setValue(dbrid, type, count, value);
    try {
    JNI._ch_arrayPut(type,count,ch.getChannelID(),dbrid);
    } finally {
      JNI._dbr_destroy(dbrid);
    }
  }

  public void ch_arrayPut(int type, int count, JNIChannel ch, Object value, PutListener l) throws CaException {
    assertThread();
    JNIPutCallback callback= new JNIPutCallback(ch, l);
    int dbrid= JNI._dbr_create(type,count);
    JNI._dbr_setValue(dbrid, type, count, value);
    try {
      JNI._ch_arrayPutCallback(type,count,ch.getChannelID(),dbrid, callback);
    } catch(CaException cae) {
      throw cae;
    } finally {
      JNI._dbr_destroy(dbrid);
    }
  }



  protected LinkedList _GetQueue= new LinkedList();

  public void ch_arrayGet(DBR dbr, JNIChannel ch) throws CaException {
    assertThread();
    int type= dbr.getType();
    int count= dbr.getCount();
    int dbrid= JNI._dbr_create(type,count);

    synchronized(_GetQueue) {
      _GetQueue.addLast(new Get(dbr,dbrid));
      JNI._ch_arrayGet(type,count,ch.getChannelID(),dbrid);
    }
  }

  public void ch_arrayGet(int type, int count, JNIChannel ch, GetListener l) throws CaException {
    assertThread();
    JNIGetCallback callback= new JNIGetCallback(ch,l);
    JNI._ch_arrayGetCallback(type,count,ch.getChannelID(),callback);
  }


  protected void clearGetQueue() {
    for (Iterator it= _GetQueue.iterator(); it.hasNext();) {
      ((Get)it.next()).update();
    }
    _GetQueue.clear();
  }


  public JNIMonitor ch_addMonitor(int type, int count, JNIChannel ch, MonitorListener l, int mask) throws CaException {
    assertThread();
    JNIMonitorCallback callback= new JNIMonitorCallback(ch,l);
    JNIMonitor res= new JNIMonitor(this,type,count,ch,l,mask);
    res.setMonitorID(JNI._ch_addMonitor(type,count,ch.getChannelID(),callback,mask));
    return res;
  }

  public void ch_clearMonitor(JNIMonitor monitor) throws CaException {
    assertThread();
    JNI._ch_clearMonitor(monitor.getMonitorID());
  }

  protected void assertThread() {
    synchronized(_threadList) {
      for(Iterator it= _threadList.iterator(); it.hasNext(); ) {
        if(Thread.currentThread() ==  (Thread)it.next()) return;
      }
    }
    throw new IllegalStateException("Illegal calling thread");
  }

  class Get {
    DBR dbr;
    int dbrid;
    public Get(DBR dbr, int dbrid) {
      this.dbr= dbr;
      this.dbrid= dbrid;
    }
    public void update() {
      JNI._dbr_update(dbr, dbrid);
      JNI._dbr_destroy(dbrid);
    }
  }




}





