/**********************************************************************
 * 
 *      Original Author: Eric Boucher
 *      Date:            05/05/2003
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 *      Beamline Controls & Data Acquisition Group
 *      Experimental Facilities Division
 *      Advanced Photon Source
 *      Argonne National Laboratory
 *
 *
 * JNIContextTCI.java,v 1.1.1.1 2003/09/11 14:42:49 boucher Exp
 *
 * Modification Log:
 * 01. 05/07/2003  erb  initial development
 *
 */


package gov.aps.epics.jca.jni;

import gov.aps.epics.jca.*;
import gov.aps.epics.jca.dbr.*;

import java.util.LinkedList;
import java.util.Iterator;



public class JNIContextTCI extends JNIContext implements Runnable {

  static protected int _num;


  protected Thread _thread;
  protected int    _callback;
  protected String _name;


  public JNIContextTCI(String name, int callback) throws CaException {
    
    if(name==null || name.trim().equals("")) {
      do {
        name="JNIContextTCI-"+(_num++);
      } while(lookupContext(name)!=null);
    } else {
      if(lookupContext(name)!=null) throw new IllegalArgumentException("Context "+name+" already exists");
    }
    _name= name;
    

    setCtxtID( ctxt_contextCreate(callback) );
  
    setContextMessageListener(new DefaultContextMessageListener());
    setContextExceptionListener(new DefaultContextExceptionListener());
  
    registerContext(this);
    
  }



  public int ctxt_getThreadCtrlMode() {
    return THREAD_CTRL_INTERNAL;
  }

  public int ctxt_getCallbackMode() {
    return _callback;
  }

  public String ctxt_getName() {
    return _thread.getName();
  }
  
  public int ctxt_contextCreate(int callback) throws CaException {
    if(_thread!=null) throw new IllegalStateException("Context already created");
    _thread= new Thread(this,_name);
    _thread.setDaemon(true);
    _thread.start();
    _callback=callback;
    processRequest(new ContextCreateRequest());
    return 0;
  }

  public void ctxt_contextDestroy() throws CaException {
    assertValidity();
    processRequest(new ContextDestroyRequest());
    unregisterContext(this);
    _thread=null;
    _keepRunning=false;
  }

  public void ctxt_setMessageListener(ContextMessageListener l) throws CaException {
    assertValidity();
    processRequest(new MessageListenerRequest(l));
  }

  public void ctxt_setExceptionListener(ContextExceptionListener l) throws CaException {
    assertValidity();
    processRequest(new ExceptionListenerRequest(l));
  }

  public void ctxt_pendIO(double timeout) throws CaException {
    assertValidity();
    processRequest(new PendIORequest(timeout));
    
  }

  public boolean  ctxt_testIO() throws CaException {
    assertValidity();
    return ((Boolean)processRequest(new TestIORequest())).booleanValue();
  }

  public void ctxt_pendEvent(double time) throws CaException {
    assertValidity();
    processRequest(new PendEventRequest(time));
  }

  public void ctxt_poll() throws CaException {
    assertValidity();
    processRequest(new PollRequest());
  }

  public void ctxt_flushIO() throws CaException {
    assertValidity();
    processRequest(new FlushIORequest());
  }

  public void ctxt_attachThread() throws CaException {
    assertValidity();
  }


  public JNIChannel ch_channelCreate(String name, ConnectionListener l) throws CaException {
    assertValidity();
    return (JNIChannel)processRequest(new ChannelCreateRequest(name,l));
  }

  public void ch_channelDestroy(JNIChannel ch) throws CaException {
    assertValidity();
    processRequest(new ChannelDestroyRequest(ch));
  }

  public void ch_setConnectionListener(JNIChannel ch, ConnectionListener l) throws CaException {
    assertValidity();
    processRequest(new ConnectionListenerRequest(ch, l));
  }

  public void ch_setAccessRightsListener(JNIChannel ch, AccessRightsListener l) throws CaException {
    assertValidity();
    processRequest(new AccessRightsRequest(ch,l));
  }

  public int ch_getFieldType(JNIChannel ch) {
    assertValidity();
    return ((Integer)processRequestSafe(new GetFieldTypeRequest(ch))).intValue();
  }

  public int ch_getElementCount(JNIChannel ch) {
    assertValidity();
    return ((Integer)processRequestSafe(new GetElementCountRequest(ch))).intValue();
  }

  public int ch_getState(JNIChannel ch) {
    assertValidity();
    return ((Integer)processRequestSafe(new GetStateRequest(ch))).intValue();
  }

  public String ch_getHostName(JNIChannel ch) {
    assertValidity();
    return (String)processRequestSafe(new GetHostNameRequest(ch));
  }

  public boolean ch_getReadAccess(JNIChannel ch) {
    assertValidity();
    return ((Boolean)processRequestSafe(new GetReadAccessRequest(ch))).booleanValue();
  }

  public boolean ch_getWriteAccess(JNIChannel ch) {
    assertValidity();
    return ((Boolean)processRequestSafe(new GetWriteAccessRequest(ch))).booleanValue();
  }

  public void ch_arrayPut(int type, int count, JNIChannel ch, Object value) throws CaException {
    assertValidity();
    processRequest(new ArrayPutRequest(type,count,ch,value));
  }

  public void ch_arrayPut(int type, int count, JNIChannel ch, Object value, PutListener l) throws CaException {
    assertValidity();
    processRequest(new ArrayPutCallbackRequest(type,count,ch,value,l));
  }


  public void ch_arrayGet(DBR dbr, JNIChannel ch) throws CaException {
    assertValidity();
    processRequest(new ArrayGetRequest(dbr,ch));
  }

  public void ch_arrayGet(int type, int count, JNIChannel ch, GetListener l) throws CaException {
    assertValidity();
    processRequest(new ArrayGetCallbackRequest(type,count,ch,l));
  }


  public JNIMonitor ch_addMonitor(int type, int count, JNIChannel ch, MonitorListener l, int mask) throws CaException {
    assertValidity();
    return (JNIMonitor)processRequest(new AddMonitorRequest(type,count,ch,l,mask));
  }

  public void ch_clearMonitor(JNIMonitor monitor) throws CaException {
    assertValidity();
    processRequest(new ClearMonitorRequest(monitor));
  }

  protected void assertValidity() {
    if(_thread==null) throw new IllegalStateException("Invalid context");
  }


  // CA THREAD 

  protected LinkedList _requestQueue= new LinkedList();
  protected boolean    _keepRunning= true;

  public void run() {
    while (_keepRunning) {
      if (_requestQueue.isEmpty()) Thread.yield();
      else {
        Request r= (Request) _requestQueue.removeFirst();
        synchronized(r) {
          try {
            r.process();
          } catch(CaException cae) {
            r.setException(cae);
          }
          r.notifyAll();
        }
      }
    }
  }

  protected Object processRequestSafe(Request r) {
    try {
      return processRequest(r);
    } catch(CaException cae) {
    }
    return null;
  }

  protected Object processRequest(Request r) throws CaException {
    synchronized(r) {
      _requestQueue.addLast(r);
      try {
        r.wait();
      } catch (InterruptedException ie) {
      }
      r.fireException();
      return r.result();
    }
  }
  
  
  // INTERNAL SERVICES


  LinkedList _GetQueue= new LinkedList();

  class Get {
    DBR dbr;
    int dbrid;
    public Get(DBR dbr, int dbrid) {
      this.dbr= dbr;
      this.dbrid= dbrid;
    }
    public void update() {
      JNI._dbr_update(dbr, dbrid);
      JNI._dbr_destroy(dbrid);
    }
  }

  void clearGetQueue() {
    for (Iterator it= _GetQueue.iterator(); it.hasNext();) {
      ((Get)it.next()).update();
    }
    _GetQueue.clear();
  }



  JNIContext getJNIContext() {
    return this;
  }


  class Request {
    protected CaException _cause;

    public void process() throws CaException {
    }
    public Object result() {
      return null;
    }
    public void setException(CaException cause) {
      _cause=cause;
    }
    public void fireException() throws CaException {
      if(_cause!=null) throw new CaException(_cause);
    }
  }


  class ContextCreateRequest extends Request {
    public void process() throws CaException {
      JNI._ctxt_contextCreate(ctxt_getCallbackMode());
    }
  }

  class ContextDestroyRequest extends Request {
    public void process() throws CaException {
      JNI._ctxt_contextDestroy();
    }
  }

  class MessageListenerRequest extends Request {
    protected ContextMessageListener _l;
    public MessageListenerRequest(ContextMessageListener l) {
      _l=l;
    }
    public void process() throws CaException {
      JNIContextMessageCallback cb=null;
      if(_l!=null) cb= new JNIContextMessageCallback(getJNIContext(),_l);
      JNI._ctxt_setMessageCallback(cb);
    }
  }
  
  class ExceptionListenerRequest extends Request {
    protected ContextExceptionListener _l;
    public ExceptionListenerRequest(ContextExceptionListener l) {
      _l=l;
    }
    public void process() throws CaException {
      JNIContextExceptionCallback cb=null;
      if(_l!=null) cb= new JNIContextExceptionCallback(getJNIContext(),_l);
      JNI._ctxt_setExceptionCallback(cb);
    }
  }
  
  class PendIORequest extends Request {
    protected double _timeout;
    public PendIORequest(double timeout) {
      _timeout= timeout;
    }
    public void process() throws CaException {
      JNI._ctxt_pendIO(_timeout);
      clearGetQueue();
    }
  }

  class TestIORequest extends Request {
    protected boolean _res;
    public void process() throws CaException {
      _res= JNI._ctxt_testIO();
    }
    public Object result() {
      return new Boolean(_res);
    }
  }

  class PendEventRequest extends Request {
    protected double _timeout;
    public PendEventRequest(double timeout) {
      _timeout= timeout;
    }
    public void process() throws CaException {
      JNI._ctxt_pendEvent(_timeout);
      clearGetQueue();
    }
  }

  class PollRequest extends Request {
    public void process() throws CaException {
      JNI._ctxt_poll();
    }
  }

  class FlushIORequest extends Request {
    public void process() throws CaException {
      JNI._ctxt_flushIO();
    }
  }


  class ChannelCreateRequest extends Request {
    protected JNIChannel _res;
    protected String _name;
    protected ConnectionListener _l;
    public ChannelCreateRequest(String name, ConnectionListener l) {
      _res= new JNIChannel(getJNIContext(), name,l);
      _name= name;
      _l=l;
    }
    public void process() throws CaException {
      JNIConnectionCallback cb=null;
      if(_l!=null) cb= new JNIConnectionCallback(_res, _l);
      _res.setChannelID(JNI._ch_channelCreate(_name,cb));
    }
    public Object result() {
      return _res;
    }
  }

  class ChannelDestroyRequest extends Request {
    protected JNIChannel _ch;
    public ChannelDestroyRequest(JNIChannel ch) {
      _ch= ch;
    }
    public void process() throws CaException {
      JNI._ch_channelDestroy(_ch.getChannelID());
    }
  }

  class ConnectionListenerRequest extends Request {
    protected JNIChannel _ch;
    protected ConnectionListener _l;
    public ConnectionListenerRequest(JNIChannel ch, ConnectionListener l) {
      _ch= ch;
      _l=l;
    }
    public void process() throws CaException {
      JNIConnectionCallback cb=null;
      if(_l!=null) cb= new JNIConnectionCallback(_ch,_l);
      JNI._ch_setConnectionCallback(_ch.getChannelID(), cb);
    }
  }

  
    class AccessRightsRequest extends Request {
      protected JNIChannel _ch;
      protected AccessRightsListener _l;
      public AccessRightsRequest(JNIChannel ch, AccessRightsListener l) {
        _ch= ch;
        _l=l;
      }
      public void process() throws CaException {
        JNIAccessRightsCallback cb=null;
        if(_l!=null) cb= new JNIAccessRightsCallback(_ch,_l);
        JNI._ch_setAccessRightsCallback(_ch.getChannelID(), cb);
      }
    }


  class GetFieldTypeRequest extends Request {
    protected int _res;
    protected JNIChannel _ch;
    public GetFieldTypeRequest(JNIChannel ch) { _ch= ch; }
    public void process() throws CaException { _res= JNI._ch_getFieldType(_ch.getChannelID()); }
    public Object result() { return new Integer(_res); }
  }

  class GetElementCountRequest extends Request {
    protected int _res;
    protected JNIChannel _ch;
    public GetElementCountRequest(JNIChannel ch) { _ch= ch; }
    public void process() throws CaException { _res= JNI._ch_getElementCount(_ch.getChannelID()); }
    public Object result() { return new Integer(_res); }
  }
  
  class GetStateRequest extends Request {
    protected int _res;
    protected JNIChannel _ch;
    public GetStateRequest(JNIChannel ch) { _ch= ch; }
    public void process() throws CaException { _res= JNI._ch_getState(_ch.getChannelID()); }
    public Object result() { return new Integer(_res); }
  }
  
  class GetHostNameRequest extends Request {
    protected String _res;
    protected JNIChannel _ch;
    public GetHostNameRequest(JNIChannel ch) { _ch= ch; }
    public void process() throws CaException { _res= JNI._ch_getHostName(_ch.getChannelID()); }
    public Object result() { return _res; }
  }
  
  class GetReadAccessRequest extends Request {
    protected boolean _res;
    protected JNIChannel _ch;
    public GetReadAccessRequest(JNIChannel ch) { _ch= ch; }
    public void process() throws CaException { _res= JNI._ch_getReadAccess(_ch.getChannelID()); }
    public Object result() { return new Boolean(_res); }
  }
  
  class GetWriteAccessRequest extends Request {
    protected boolean _res;
    protected JNIChannel _ch;
    public GetWriteAccessRequest(JNIChannel ch) { _ch= ch; }
    public void process() throws CaException { _res= JNI._ch_getWriteAccess(_ch.getChannelID()); }
    public Object result() { return new Boolean(_res); }
  }



  class ArrayPutRequest extends Request {
    protected int _type,_count;
    protected JNIChannel _ch;
    protected Object _value;
    public ArrayPutRequest(int type, int count, JNIChannel ch, Object value) {
      _type=type;_count=count;_ch=ch;_value=value;
    }
    public void process() throws CaException {
      int dbrid= JNI._dbr_create(_type,_count);
      JNI._dbr_setValue(dbrid, _type, _count, _value);
      try {
      JNI._ch_arrayPut(_type,_count,_ch.getChannelID(),dbrid);
      } finally {
        JNI._dbr_destroy(dbrid);
      }
    }
  }

  class ArrayPutCallbackRequest extends Request {
    protected int _type,_count;
    protected JNIChannel _ch;
    protected Object _value;
    protected PutListener _l;
    public ArrayPutCallbackRequest(int type, int count, JNIChannel ch, Object value, PutListener l) {
      _type=type;_count=count;_ch=ch;_value=value;_l=l;
    }
    public void process() throws CaException {
      JNIPutCallback callback= new JNIPutCallback(_ch, _l);
      int dbrid= JNI._dbr_create(_type,_count);
      JNI._dbr_setValue(dbrid, _type, _count, _value);
      try {
      JNI._ch_arrayPutCallback(_type,_count,_ch.getChannelID(),dbrid, callback);
      } finally {
      JNI._dbr_destroy(dbrid);
      }
    }
  }



  class ArrayGetRequest extends Request {
    protected DBR _dbr;
    protected JNIChannel _ch;
    public ArrayGetRequest(DBR dbr, JNIChannel ch) {
      _dbr= dbr;
      _ch=ch;
    }
    public void process() throws CaException {
      int type= _dbr.getType();
      int count= _dbr.getCount();
      int dbrid= JNI._dbr_create(type,count);

      _GetQueue.addLast(new Get(_dbr,dbrid));
      JNI._ch_arrayGet(type,count,_ch.getChannelID(),dbrid);
    }
  }


  class ArrayGetCallbackRequest extends Request {
    protected int _type,_count;
    protected JNIChannel _ch;
    protected GetListener _l;
    public ArrayGetCallbackRequest(int type, int count, JNIChannel ch, GetListener l) {
      _type=type;_count=count;_ch=ch;_l=l;
    }
    public void process() throws CaException {
      JNIGetCallback callback= new JNIGetCallback(_ch,_l);
      JNI._ch_arrayGetCallback(_type,_count,_ch.getChannelID(),callback);
    }
  }

  class AddMonitorRequest extends Request {
    protected int _type;
    protected int _count;
    protected JNIChannel _ch;
    protected MonitorListener _l;
    protected int _mask;
    protected JNIMonitor _res;
    public AddMonitorRequest(int type, int count, JNIChannel ch, MonitorListener l, int mask) {
      _type=type;
      _count=count;
      _ch=ch;
      _l=l;
      _mask=mask;
      _res= new JNIMonitor(getJNIContext(),type,count,ch,l,mask);
    }
    public void process() throws CaException {
      JNIMonitorCallback callback= new JNIMonitorCallback(_ch,_l);
      _res.setMonitorID(JNI._ch_addMonitor(_type,_count,_ch.getChannelID(),callback,_mask));
    }
    public Object result() {
      return _res;
    }
  }


  class ClearMonitorRequest extends Request {
    protected JNIMonitor _m;
    public ClearMonitorRequest(JNIMonitor m) {
      _m=m;
    }
    public void process() throws CaException {
      JNI._ch_clearMonitor(_m.getMonitorID());
    }
  }


}





