/**********************************************************************
 * 
 *      Original Author: Eric Boucher
 *      Date:            05/05/2003
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 *      Beamline Controls & Data Acquisition Group
 *      Experimental Facilities Division
 *      Advanced Photon Source
 *      Argonne National Laboratory
 *
 *
 * Ca.java,v 1.1.1.1 2003/09/11 14:42:49 boucher Exp
 *
 * Modification Log:
 * 01. 05/07/2003  erb  initial development
 *
 */

package jca;

import jca.event.*;
import jca.dbr.*;
import java.util.*;




/**
 * The Java Channel Access class.
 * 
 * @author Eric BOUCHER
 * @see jca.PV
 * @see jca.Monitor
 *
 * @deprecated  Use gov.apsjca.Ca instead. 
 */
public final class Ca {

  /*
   * Channel Access OP code. Get request.
   */
  public static final int OP_GET        = 0;
  /*
   * Channel Access OP code. Put request.
   */
  public static final int OP_PUT        = 1;
  /*
   * Channel Access OP code. Search request.
   */
  public static final int OP_SEARCH     = 2;
  /*
   * Channel Access OP code. Add event request.
   */
  public static final int OP_ADD_EVENT  = 3;
  /*
   * Channel Access OP code. Clear event request.
   */
  public static final int OP_CLEAR_EVENT= 4;
  /*
   * Channel Access OP code. Other request.
   */
  public static final int OP_OTHER      = 5;
  /*
   * Channel Access OP code. Connection up notification.
   */
  public static final int OP_CONN_UP    = 6;
  /*
   * Channel Access OP code. Connection down notification.
   */
  public static final int OP_CONN_DOWN  = 7;


  static protected gov.aps.epics.jca.Context _theContext;

  /**
   * Initializes the Channel Access threads.
   *
   */
  public static synchronized void init() {
    init(Thread.NORM_PRIORITY+2, Thread.NORM_PRIORITY+1);
  }

  public static synchronized void init(int requestPriority, int eventPriority) {
    try {
    _theContext= gov.aps.epics.jca.Ca.createContext();
    } catch(gov.aps.epics.jca.CaException cae) {
      _theContext=null;
    }
  }

  /**
   * Frees all CA resources and exits the Channel Access thread
   */
  public static void exit() {
    if(_theContext!=null) {
      try {
        _theContext.destroy();
      }catch(gov.aps.epics.jca.CaException cae) {
      }
    }
    _theContext=null;
  }

  /**
   * Sends all buffered IO requests. Usually performed by pendIO or pendEvent.
   * However, some users may wish to flush remote operation requests early so
   * that they can perform operations in parallel with outstanding remote
   * operations.
   *
   * @see       jca.Ca#pendEvent
   * @see       jca.Ca#pendIO
   */
  public static void flushIO() {
    if(_theContext==null) return;
    try {
      _theContext.flushIO();
    } catch(gov.aps.epics.jca.CaException cae) {
    }
  }

  /**
   * Flushes the send buffer and waits until outstanding queries complete.
   * At this time queries include calls to PV(name) and PV.get(value). 
   * <p>
   * Comments: Any values written into your program's variables by a CA synchronous
   * request should not be referenced by your program until Ca.pendIO() has been called.
   *
   * @param  timeout    time in seconds to wait before the function times out. A timeout 
   *                    of zero means wait forever.
   * @exception jca.TimeOutException    The timeout has expire and all outstanding queries
   *                                    were not completed.
   */
  public static void pendIO(double timeout) throws jca.TimeOutException {
    if(_theContext==null) return;
    synchronized(_queue) {
    try {
      _theContext.pendIO(timeout);
    } catch(gov.aps.epics.jca.CaException cae) {
      if(cae.getCode()==gov.aps.epics.jca.CaConstants.ECA_TIMEOUT) throw new TimeOutException();
    } finally {
      clearQueue();
    }
    }
  }

  /**
   * Flushes the send buffer and waits for asynchronous event. Asynchronous event
   * include all call to PV(name,listener), PV.get(value,listener), PV.put(value,listener)
   * and all monitors events.
   * This routine will not return before the period expires and all unfinished 
   * channel access labor has been processed.
   * <p>
   * Comments: A very short timeout such as 0.0001 seconds will result in a poll
   *           (i.e. channel access will just check for file descriptor activity
   *           and then return).
   *
   * @param  period     time in seconds to wait before the function returns. A period 
   *                    of zero means wait forever.
   */
  public static void pendEvent(double period) {
    if(_theContext==null) return;
    synchronized(_queue) {
    try {
      _theContext.pendEvent(period);
    } catch(gov.aps.epics.jca.CaException cae) {
    } finally {
      clearQueue();
    }
    }
  }

  static class DBRAdapter {
    DBR _old;
    gov.aps.epics.jca.dbr.DBR _new;
    public DBRAdapter(DBR dbrOld, gov.aps.epics.jca.dbr.DBR dbrNew) {
      _old= dbrOld;
      _new=dbrNew;
    }
    public void update() {
      _old.update(_new);
    }
  }

  static protected Vector _queue= new Vector();

  static protected void clearQueue() {
    synchronized(_queue) {
      DBRAdapter[] dbr= (DBRAdapter[])_queue.toArray(new DBRAdapter[0]);

      for(int t=0; t<dbr.length; ++t) {
        dbr[t].update();
      }

      _queue.clear();
    }
  }

  static protected void get(PV pv, DBR ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.GetFailException {
    pv.assertValidity();
    synchronized(_queue) {
      try {
      gov.aps.epics.jca.dbr.DBR dbr= pv.m_channel.get(ivalue.type(), ivalue.count());

      DBRAdapter dbra= new DBRAdapter(ivalue, dbr);
      _queue.add(dbra);
      } catch ( gov.aps.epics.jca.CaException cae) {
        switch(cae.getCode()) {
          case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
          case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
          case gov.aps.epics.jca.CaConstants.ECA_GETFAIL : throw new GetFailException();
        }
      }
    }
  }


  static protected gov.aps.epics.jca.Context getContext() {
    if(_theContext==null) throw new IllegalStateException("CA has not been initialized");
    return _theContext;
  }


  protected void finalize() throws Throwable { 
    exit();
    super.finalize();
  }

}



