/**********************************************************************
 * 
 *      Original Author: Eric Boucher
 *      Date:            05/05/2003
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contract
 *      W-31-109-ENG-38 at Argonne National Laboratory.
 *
 *      Beamline Controls & Data Acquisition Group
 *      Experimental Facilities Division
 *      Advanced Photon Source
 *      Argonne National Laboratory
 *
 *
 * PV.java,v 1.1.1.1 2003/09/11 14:42:49 boucher Exp
 *
 * Modification Log:
 * 01. 05/07/2003  erb  initial development
 *
 */


package jca;

import jca.dbr.*;
import jca.event.*;

/**
 * This class define an object to access a channel access process variable.
 *
 * @see         Ca
 * @see         Monitor
 *
 * @deprecated
 */
public class PV extends Object {
  public static final int NEVER_CONN =0;
  public static final int PREV_CONN  =1;
  public static final int CONN       =2;
  public static final int CLOSED     =3;

  public static final int DBE_VALUE  =1;
  public static final int DBE_LOG    =2;
  public static final int DBE_ALARM  =4;
  
  protected String                m_name="";
  protected Object                m_puser= null;
  protected gov.aps.epics.jca.Channel   m_channel=null;
  
  /**
   * Creates a PV and connect it to the specified channel access process variable.
   * The connection request is queued and not forwarded to the IOC until one of 
   * Ca.flushIO, Ca.pendIO or Ca.pendEvent function is called.
   *
   * @param     name    The name of the process variable to connect to.
   */
  public PV(String name) {
    this(name,null,null);
  }

  /**
   * Creates a PV and connect it to the specified channel access process variable.
   * The connection request is queued and not forwarded to the IOC until one of 
   * Ca.flushIO, Ca.pendIO or Ca.pendEvent function is called.
   *
   * @param     name    The name of the process variable to connect to.
   * @param     puser   A user defined object to attach to the pv.
   */
  public PV(String name, Object puser) {
    this(name,puser,null);
  }

  class ConnectionAdapter implements gov.aps.epics.jca.ConnectionListener {
    PV _pv;
    ConnectionListener _l;
    ConnectionAdapter(PV pv, ConnectionListener l) {
      _pv=pv;
      _l=l;
    }
    public void connectionChanged(gov.aps.epics.jca.ConnectionEvent ev) {
      _l.connectionChanged(new ConnectionEvent(_pv, (ev.isConnected()?Ca.OP_CONN_UP:Ca.OP_CONN_DOWN)));
    }
  }

  /**
   * Creates a PV and connect it to the specified channel access process variable.
   * The connection request is queued and not forwarded to the IOC until one of 
   * Ca.flushIO, Ca.pendIO or Ca.pendEvent function is called.
   *
   * @param     name    The name of the process variable to connect to.
   * @param     puser   A user defined object to attach to the pv.
   * @param     l       A ConnectionListener to notify whenever the connection state
   *                    of the PV change.
   */
  public PV(String name, Object puser, jca.event.ConnectionListener l) {
    m_name = new String(name);
    m_puser= puser;

    ConnectionAdapter cl=null;
    if(l!=null) cl= new ConnectionAdapter(this,l);

    try {
      m_channel= Ca.getContext().createChannel(m_name,cl);
    }catch(gov.aps.epics.jca.CaException cae) {
      m_channel=null;
    }
  }


  ConnectionAdapter _cl=null;

  /**
   * Changes the Connectionlistener attached to the PV.
   * 
   * @param     l       the Connectionlistener to be notify of state change.
   *                    Set it to null to avoid being notified.
   */
  public void setConnectionListener(jca.event.ConnectionListener l) 
    throws BadPVException  {
    assertValidity();
    
    try {
      m_channel.removeConnectionListener(_cl);
      _cl= (l==null?null:new ConnectionAdapter(this,l));
      m_channel.addConnectionListener(_cl);
    } catch(gov.aps.epics.jca.CaException cae) {
    }
  }

  /**
   * Disconnect the PV from the channel access variable and clear all the events associated
   * with this PV.
   */
  public void clear() 
    throws jca.BadPVException {
    assertValidity();
    try {
    m_channel.destroy();
    } catch(gov.aps.epics.jca.CaException cae) {
    }
  }

  /**
   * Return the PV id. Each PV has its own id.
   */
  public int id() {
    return (m_channel==null?0:1);
  }

  /**
   * Return the native type of the PV.
   */
  public int fieldType()
    throws jca.BadPVException {
    assertValidity();
    return m_channel.getFieldType();
  }

  /**
   * Return the native element count of the PV.
   */
  public int elementCount()
    throws jca.BadPVException {
    assertValidity();
    return m_channel.getElementCount();
  }

  /**
   * Return the name of the PV.
   */
  public String name()
    throws jca.BadPVException {
    assertValidity();
    return m_channel.getName();
  }

  /**
   * Return the user Object associated with this PV.
   */
  public Object puser()
    throws jca.BadPVException {
    assertValidity();
    return m_puser;
  }

  /**
   * Change the user Object associated with this PV.
   */
  public void puser(Object arg)
    throws jca.BadPVException {
    assertValidity();
    m_puser= arg;
  }

  /**
   * Return the state of the PV.
   */
  public int state()
    throws jca.BadPVException {
    assertValidity();
    return m_channel.getState();
  }

  /**
   * Return the host name of the PV.
   */
  public String hostName()
    throws jca.BadPVException {
    assertValidity();
    return m_channel.getHostName();
  }

  /**
   * Return true if the PV has read access.
   */
  public boolean readAccess()
    throws jca.BadPVException {
    assertValidity();
    return m_channel.getReadAccess();
  }

  /**
   * Return true if the PV has write access.
   */
  public boolean writeAccess()
    throws jca.BadPVException {
    assertValidity();
    return m_channel.getWriteAccess();
  }

  protected void assertValidity() {
    if(m_channel==null) throw new BadPVException();
  }

  /**
   * Synchronously put a DBR object into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The DBR Objectto send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.dbr.DBR
   */
  public void put(DBR ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {

    if(ivalue instanceof DBR_DBString) {
      put(new String[] { ((DBR_DBString)ivalue).value() } );
    } else if(ivalue instanceof DBR_DBShort) {
      put( ((DBR_DBShort)ivalue).value()  );
    } else if(ivalue instanceof DBR_DBFloat) {
      put( ((DBR_DBFloat)ivalue).value()  );
    } else if(ivalue instanceof DBR_DBEnum) {
      put( ((DBR_DBEnum)ivalue).value()  );
    } else if(ivalue instanceof DBR_DBByte) {
      put( ((DBR_DBByte)ivalue).value()  );
    } else if(ivalue instanceof DBR_DBInt) {
      put( ((DBR_DBInt)ivalue).value()  );
    } else if(ivalue instanceof DBR_DBDouble) {
      put( ((DBR_DBDouble)ivalue).value()  );
    }
  }

  /**
   * Synchronously put a String into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The String to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(String ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new String[] { ivalue });
  }

  public void put(String[] ivalue) throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue);
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }

  /**
   * Synchronously put a short value into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The short value to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(short ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    
    put(new short[] {ivalue});
  }

  /**
   * Synchronously put an array of short into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The array of short to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(short[] ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue);
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }

  /**
   * Synchronously put a float value into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The float value to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(float ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new float[] { ivalue });
  }

  /**
   * Synchronously put an array of float into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The array of float to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(float[] ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue);
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }

  /**
   * Synchronously put a byte value into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The byte value to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(byte ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new byte[] { ivalue } );
  }

  /**
   * Synchronously put an array of byte into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The array of byte to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(byte[] ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue);
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }

  /**
   * Synchronously put an int value into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The int value to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(int ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new int[] { ivalue } );
  }

  /**
   * Synchronously put an array of int into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The array of int to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(int[] ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue);
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }

  /**
   * Synchronously put a double value into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The double value to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(double ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new double[] { ivalue });
  }

  /**
   * Synchronously put an array of double into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The array of double to send.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   */
  public void put(double[] ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue);
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }

  class PutAdapter implements gov.aps.epics.jca.PutListener {
    protected PV _pv;
    protected PutListener _l;
    protected Object _arg;

    public PutAdapter(PV pv, PutListener l, Object arg) {
      _pv= pv;
      _l=l;
      _arg=arg;
    }
    public void putCompleted(gov.aps.epics.jca.PutEvent ev) {
      _l.putCompleted(new PutEvent(_pv,ev.getType(), ev.getCount(),_arg,ev.getStatus()));
    }
  }

  /**
   * Asynchronously put a DBR object into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The DBR Object to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.dbr.DBR
   * @see       jca.event.PutListener
   */
  public void put(DBR ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    
    if(ivalue instanceof DBR_DBString) {
      put(new String[] { ((DBR_DBString)ivalue).value() }, l, arg );
    } else if(ivalue instanceof DBR_DBShort) {
      put( ((DBR_DBShort)ivalue).value(), l, arg  );
    } else if(ivalue instanceof DBR_DBFloat) {
      put( ((DBR_DBFloat)ivalue).value(), l, arg  );
    } else if(ivalue instanceof DBR_DBEnum) {
      put( ((DBR_DBEnum)ivalue).value(), l, arg  );
    } else if(ivalue instanceof DBR_DBByte) {
      put( ((DBR_DBByte)ivalue).value(), l, arg  );
    } else if(ivalue instanceof DBR_DBInt) {
      put( ((DBR_DBInt)ivalue).value(), l, arg  );
    } else if(ivalue instanceof DBR_DBDouble) {
      put( ((DBR_DBDouble)ivalue).value(), l, arg  );
    }
    
  }

  /**
   * Asynchronously put a String into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The String to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(String ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new String[] {ivalue},l,arg);
  }

  public void put(String[] ivalue, PutListener l, Object arg) throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue, new PutAdapter(this,l,arg));
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }


  /**
   * Asynchronously put a short into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The short value to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(short ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new short[] { ivalue }, l, arg);
  }

  /**
   * Asynchronously put an array of short into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The array of short to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(short[] ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue, new PutAdapter(this,l,arg));
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }

  /**
   * Asynchronously put a float into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The float to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(float ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new float[] { ivalue },l,arg);
  }

  /**
   * Asynchronously put an array of float into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The array of short to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(float[] ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue, new PutAdapter(this,l,arg));
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }

  /**
   * Asynchronously put a byte into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The byte to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(byte ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new byte[] { ivalue },l,arg);
  }

  /**
   * Asynchronously put an array of byte into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The byte to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(byte[] ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue, new PutAdapter(this,l,arg));
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }

  /**
   * Asynchronously put a int into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The int to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(int ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new int[] { ivalue } ,l,arg);
  }

  /**
   * Asynchronously put an array of int into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The array of int to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(int[] ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue, new PutAdapter(this,l,arg));
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }

  /**
   * Asynchronously put a double into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The double to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(double ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    put(new double[] { ivalue },l,arg);
  }

  /**
   * Asynchronously put an array of double into the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The array of double to send.
   * @param     l                       The PutListener to notify when the put has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.PutFailException    The put operation failled.
   * @see       jca.event.PutListener
   */
  public void put(double[] ivalue, jca.event.PutListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.PutFailException {
    assertValidity();
    try {
    m_channel.put(ivalue, new PutAdapter(this,l,arg));
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_PUTFAIL : throw new PutFailException();
      }
    }
  }




  /**
   * Synchronously get a DBR Object from the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The DBR Object requested.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.GetFailException    The get operation failled.
   * @see       jca.dbr.DBR
   */
  public void get(DBR ivalue)
    throws jca.BadTypeException, jca.BadCountException, jca.GetFailException {
    Ca.get(this,ivalue);
  }

  class GetAdapter implements gov.aps.epics.jca.GetListener {
    PV _pv;
    DBR _dbr;
    GetListener _l;
    Object _arg;
    public GetAdapter(PV pv, DBR ivalue, GetListener l, Object arg) {
      _pv= pv;
      _dbr=ivalue;
      _l=l;
      _arg=arg;
    }
    public void getCompleted(gov.aps.epics.jca.GetEvent ev) {
      _dbr.update(ev.getDBR());
      _l.getCompleted(new GetEvent(_pv,_dbr,_arg,ev.getStatus()));
    }
  }

  /**
   * Asynchronously get a DBR object from the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The DBR Object requested.
   * @param     l                       The GetListener to notify when the get has completed
   * @param     arg                     An object to pass to the listener.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.GetFailException    The get operation failled.
   * @see       jca.dbr.DBR
   * @see       jca.event.GetListener
   */
  public void get(DBR ivalue, jca.event.GetListener l, Object arg)
    throws jca.BadTypeException, jca.BadCountException, jca.GetFailException {
    assertValidity();

    GetAdapter gl= new GetAdapter(this, ivalue,l,arg);

    try {
    m_channel.get(ivalue.type(), ivalue.count(), gl);
    } catch ( gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_GETFAIL : throw new GetFailException();
      }
    }
  }


  class MonitorAdapter implements gov.aps.epics.jca.MonitorListener {
    PV _pv;
    DBR _dbr;
    MonitorListener _l;
    Object _arg;

    public MonitorAdapter(PV pv, DBR dbr, MonitorListener l, Object arg) {
      _pv=pv;
      _dbr= dbr;
      _l=l;
      _arg=arg;
    }
    public void monitorChanged(gov.aps.epics.jca.MonitorEvent ev) {
      _dbr.update(ev.getDBR());
      _l.monitorChanged(new MonitorEvent(_pv,_dbr,_arg, ev.getStatus()));
    }
  }

  /**
   * Add a monitor to the PV.
   * The request is queued and not forwarded to the IOC until one of Ca.flushIO, Ca.pendIO,
   * Ca.pendEvent is called.
   * 
   * @param     ivalue                  The DBR Object requested.
   * @param     l                       The GetListener to notify when the get has completed
   * @param     arg                     An object to pass to the listener.
   * @param     mask                    The mask of the monitor. It can be a combination of
   *                                    Monitor.VALUE, Monitor.LOG, Monitor.ALARM.
   * @return                            A Monitor object that can be used to clear the monitor.
   * @exception jca.BadTypeException    The DBR is of an invalid type for the PV.
   * @exception jca.BadCountException   The DBR has an invalid element count for the PV.
   * @exception jca.MonitorFailException The add monitor operation failled.
   * @see       jca.dbr.DBR
   * @see       jca.event.GetListener
   * @see       jca.Monitor
   */
  public jca.Monitor addMonitor(DBR ivalue, jca.event.MonitorListener l, Object arg,
                       int mask)
    throws jca.BadTypeException, jca.BadCountException, jca.MonitorFailException {
    assertValidity();

    MonitorAdapter ml= new MonitorAdapter(this, ivalue, l, arg);

    try {
      return new Monitor(m_channel.addMonitorListener(ivalue.type(), ivalue.count(), ml, mask));
    } catch(gov.aps.epics.jca.CaException cae) {
      switch(cae.getCode()) {
        case gov.aps.epics.jca.CaConstants.ECA_BADTYPE : throw new BadTypeException();
        case gov.aps.epics.jca.CaConstants.ECA_BADCOUNT: throw new BadCountException();
        case gov.aps.epics.jca.CaConstants.ECA_ADDFAIL : throw new MonitorFailException();
      }

    }
    return null;
  }


}

